//
// Copyright 2021, Sander van Harmelen
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

package gitlab

import (
	"net/http"
	"time"
)

type (
	NotesServiceInterface interface {
		ListIssueNotes(pid any, issue int64, opt *ListIssueNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error)
		GetIssueNote(pid any, issue, note int64, options ...RequestOptionFunc) (*Note, *Response, error)
		CreateIssueNote(pid any, issue int64, opt *CreateIssueNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		UpdateIssueNote(pid any, issue, note int64, opt *UpdateIssueNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		DeleteIssueNote(pid any, issue, note int64, options ...RequestOptionFunc) (*Response, error)
		ListSnippetNotes(pid any, snippet int64, opt *ListSnippetNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error)
		GetSnippetNote(pid any, snippet, note int64, options ...RequestOptionFunc) (*Note, *Response, error)
		CreateSnippetNote(pid any, snippet int64, opt *CreateSnippetNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		UpdateSnippetNote(pid any, snippet, note int64, opt *UpdateSnippetNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		DeleteSnippetNote(pid any, snippet, note int64, options ...RequestOptionFunc) (*Response, error)
		ListMergeRequestNotes(pid any, mergeRequest int64, opt *ListMergeRequestNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error)
		GetMergeRequestNote(pid any, mergeRequest, note int64, options ...RequestOptionFunc) (*Note, *Response, error)
		CreateMergeRequestNote(pid any, mergeRequest int64, opt *CreateMergeRequestNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		UpdateMergeRequestNote(pid any, mergeRequest, note int64, opt *UpdateMergeRequestNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		DeleteMergeRequestNote(pid any, mergeRequest, note int64, options ...RequestOptionFunc) (*Response, error)
		ListEpicNotes(gid any, epic int64, opt *ListEpicNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error)
		GetEpicNote(gid any, epic, note int64, options ...RequestOptionFunc) (*Note, *Response, error)
		CreateEpicNote(gid any, epic int64, opt *CreateEpicNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		UpdateEpicNote(gid any, epic, note int64, opt *UpdateEpicNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error)
		DeleteEpicNote(gid any, epic, note int64, options ...RequestOptionFunc) (*Response, error)
	}

	// NotesService handles communication with the notes related methods
	// of the GitLab API.
	//
	// GitLab API docs:
	// https://docs.gitlab.com/api/notes/
	NotesService struct {
		client *Client
	}
)

var _ NotesServiceInterface = (*NotesService)(nil)

// Note represents a GitLab note.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/
type Note struct {
	ID           int64          `json:"id"`
	Type         NoteTypeValue  `json:"type"`
	Body         string         `json:"body"`
	Attachment   string         `json:"attachment"`
	Title        string         `json:"title"`
	FileName     string         `json:"file_name"`
	Author       NoteAuthor     `json:"author"`
	System       bool           `json:"system"`
	CreatedAt    *time.Time     `json:"created_at"`
	UpdatedAt    *time.Time     `json:"updated_at"`
	ExpiresAt    *time.Time     `json:"expires_at"`
	CommitID     string         `json:"commit_id"`
	Position     *NotePosition  `json:"position"`
	NoteableID   int64          `json:"noteable_id"`
	NoteableType string         `json:"noteable_type"`
	ProjectID    int64          `json:"project_id"`
	NoteableIID  int64          `json:"noteable_iid"`
	Resolvable   bool           `json:"resolvable"`
	Resolved     bool           `json:"resolved"`
	ResolvedAt   *time.Time     `json:"resolved_at"`
	ResolvedBy   NoteResolvedBy `json:"resolved_by"`
	Internal     bool           `json:"internal"`

	// Deprecated: use Internal instead
	Confidential bool `json:"confidential"`
}

// NoteResolvedBy represents the resolver of a GitLab note.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/
type NoteResolvedBy struct {
	ID        int64  `json:"id"`
	Username  string `json:"username"`
	Email     string `json:"email"`
	Name      string `json:"name"`
	State     string `json:"state"`
	AvatarURL string `json:"avatar_url"`
	WebURL    string `json:"web_url"`
}

// NoteAuthor represents the author of a note.
type NoteAuthor struct {
	ID        int64  `json:"id"`
	Username  string `json:"username"`
	Email     string `json:"email"`
	Name      string `json:"name"`
	State     string `json:"state"`
	AvatarURL string `json:"avatar_url"`
	WebURL    string `json:"web_url"`
}

// NotePosition represents the position attributes of a note.
type NotePosition struct {
	BaseSHA      string     `json:"base_sha"`
	StartSHA     string     `json:"start_sha"`
	HeadSHA      string     `json:"head_sha"`
	PositionType string     `json:"position_type"`
	NewPath      string     `json:"new_path,omitempty"`
	NewLine      int64      `json:"new_line,omitempty"`
	OldPath      string     `json:"old_path,omitempty"`
	OldLine      int64      `json:"old_line,omitempty"`
	LineRange    *LineRange `json:"line_range,omitempty"`
}

// LineRange represents the range of a note.
type LineRange struct {
	StartRange *LinePosition `json:"start"`
	EndRange   *LinePosition `json:"end"`
}

// LinePosition represents a position in a line range.
type LinePosition struct {
	LineCode string `json:"line_code"`
	Type     string `json:"type"`
	OldLine  int64  `json:"old_line"`
	NewLine  int64  `json:"new_line"`
}

func (n Note) String() string {
	return Stringify(n)
}

// ListIssueNotesOptions represents the available ListIssueNotes() options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-project-issue-notes
type ListIssueNotesOptions struct {
	ListOptions
	OrderBy *string `url:"order_by,omitempty" json:"order_by,omitempty"`
	Sort    *string `url:"sort,omitempty" json:"sort,omitempty"`
}

// ListIssueNotes gets a list of all notes for a single issue.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-project-issue-notes
func (s *NotesService) ListIssueNotes(pid any, issue int64, opt *ListIssueNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error) {
	return do[[]*Note](s.client,
		withPath("projects/%s/issues/%d/notes", ProjectID{pid}, issue),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// GetIssueNote returns a single note for a specific project issue.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#get-single-issue-note
func (s *NotesService) GetIssueNote(pid any, issue, note int64, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withPath("projects/%s/issues/%d/notes/%d", ProjectID{pid}, issue, note),
		withRequestOpts(options...),
	)
}

// CreateIssueNoteOptions represents the available CreateIssueNote()
// options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-issue-note
type CreateIssueNoteOptions struct {
	Body      *string    `url:"body,omitempty" json:"body,omitempty"`
	CreatedAt *time.Time `url:"created_at,omitempty" json:"created_at,omitempty"`
	Internal  *bool      `url:"internal,omitempty" json:"internal,omitempty"`
}

// CreateIssueNote creates a new note to a single project issue.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-issue-note
func (s *NotesService) CreateIssueNote(pid any, issue int64, opt *CreateIssueNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPost),
		withPath("projects/%s/issues/%d/notes", ProjectID{pid}, issue),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// UpdateIssueNoteOptions represents the available UpdateIssueNote()
// options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-issue-note
type UpdateIssueNoteOptions struct {
	Body *string `url:"body,omitempty" json:"body,omitempty"`
}

// UpdateIssueNote modifies existing note of an issue.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-issue-note
func (s *NotesService) UpdateIssueNote(pid any, issue, note int64, opt *UpdateIssueNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPut),
		withPath("projects/%s/issues/%d/notes/%d", ProjectID{pid}, issue, note),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// DeleteIssueNote deletes an existing note of an issue.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#delete-an-issue-note
func (s *NotesService) DeleteIssueNote(pid any, issue, note int64, options ...RequestOptionFunc) (*Response, error) {
	_, resp, err := do[none](s.client,
		withMethod(http.MethodDelete),
		withPath("projects/%s/issues/%d/notes/%d", ProjectID{pid}, issue, note),
		withRequestOpts(options...),
	)
	return resp, err
}

// ListSnippetNotesOptions represents the available ListSnippetNotes() options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-snippet-notes
type ListSnippetNotesOptions struct {
	ListOptions
	OrderBy *string `url:"order_by,omitempty" json:"order_by,omitempty"`
	Sort    *string `url:"sort,omitempty" json:"sort,omitempty"`
}

// ListSnippetNotes gets a list of all notes for a single snippet. Snippet
// notes are comments users can post to a snippet.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-snippet-notes
func (s *NotesService) ListSnippetNotes(pid any, snippet int64, opt *ListSnippetNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error) {
	return do[[]*Note](s.client,
		withPath("projects/%s/snippets/%d/notes", ProjectID{pid}, snippet),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// GetSnippetNote returns a single note for a given snippet.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#get-single-snippet-note
func (s *NotesService) GetSnippetNote(pid any, snippet, note int64, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withPath("projects/%s/snippets/%d/notes/%d", ProjectID{pid}, snippet, note),
		withRequestOpts(options...),
	)
}

// CreateSnippetNoteOptions represents the available CreateSnippetNote()
// options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-snippet-note
type CreateSnippetNoteOptions struct {
	Body      *string    `url:"body,omitempty" json:"body,omitempty"`
	CreatedAt *time.Time `url:"created_at,omitempty" json:"created_at,omitempty"`
}

// CreateSnippetNote creates a new note for a single snippet. Snippet notes are
// comments users can post to a snippet.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-snippet-note
func (s *NotesService) CreateSnippetNote(pid any, snippet int64, opt *CreateSnippetNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPost),
		withPath("projects/%s/snippets/%d/notes", ProjectID{pid}, snippet),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// UpdateSnippetNoteOptions represents the available UpdateSnippetNote()
// options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-snippet-note
type UpdateSnippetNoteOptions struct {
	Body *string `url:"body,omitempty" json:"body,omitempty"`
}

// UpdateSnippetNote modifies existing note of a snippet.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-snippet-note
func (s *NotesService) UpdateSnippetNote(pid any, snippet, note int64, opt *UpdateSnippetNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPut),
		withPath("projects/%s/snippets/%d/notes/%d", ProjectID{pid}, snippet, note),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// DeleteSnippetNote deletes an existing note of a snippet.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#delete-a-snippet-note
func (s *NotesService) DeleteSnippetNote(pid any, snippet, note int64, options ...RequestOptionFunc) (*Response, error) {
	_, resp, err := do[none](s.client,
		withMethod(http.MethodDelete),
		withPath("projects/%s/snippets/%d/notes/%d", ProjectID{pid}, snippet, note),
		withRequestOpts(options...),
	)
	return resp, err
}

// ListMergeRequestNotesOptions represents the available ListMergeRequestNotes()
// options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-merge-request-notes
type ListMergeRequestNotesOptions struct {
	ListOptions
	OrderBy *string `url:"order_by,omitempty" json:"order_by,omitempty"`
	Sort    *string `url:"sort,omitempty" json:"sort,omitempty"`
}

// ListMergeRequestNotes gets a list of all notes for a single merge request.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-merge-request-notes
func (s *NotesService) ListMergeRequestNotes(pid any, mergeRequest int64, opt *ListMergeRequestNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error) {
	return do[[]*Note](s.client,
		withPath("projects/%s/merge_requests/%d/notes", ProjectID{pid}, mergeRequest),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// GetMergeRequestNote returns a single note for a given merge request.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#get-single-merge-request-note
func (s *NotesService) GetMergeRequestNote(pid any, mergeRequest, note int64, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withPath("projects/%s/merge_requests/%d/notes/%d", ProjectID{pid}, mergeRequest, note),
		withRequestOpts(options...),
	)
}

// CreateMergeRequestNoteOptions represents the available
// CreateMergeRequestNote() options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-merge-request-note
type CreateMergeRequestNoteOptions struct {
	Body                    *string    `url:"body,omitempty" json:"body,omitempty"`
	CreatedAt               *time.Time `url:"created_at,omitempty" json:"created_at,omitempty"`
	Internal                *bool      `url:"internal,omitempty" json:"internal,omitempty"`
	MergeRequestDiffHeadSHA *string    `url:"merge_request_diff_head_sha,omitempty" json:"merge_request_diff_head_sha,omitempty"`
}

// CreateMergeRequestNote creates a new note for a single merge request.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-merge-request-note
func (s *NotesService) CreateMergeRequestNote(pid any, mergeRequest int64, opt *CreateMergeRequestNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPost),
		withPath("projects/%s/merge_requests/%d/notes", ProjectID{pid}, mergeRequest),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// UpdateMergeRequestNoteOptions represents the available
// UpdateMergeRequestNote() options.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-merge-request-note
type UpdateMergeRequestNoteOptions struct {
	Body *string `url:"body,omitempty" json:"body,omitempty"`
}

// UpdateMergeRequestNote modifies existing note of a merge request.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-merge-request-note
func (s *NotesService) UpdateMergeRequestNote(pid any, mergeRequest, note int64, opt *UpdateMergeRequestNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPut),
		withPath("projects/%s/merge_requests/%d/notes/%d", ProjectID{pid}, mergeRequest, note),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// DeleteMergeRequestNote deletes an existing note of a merge request.
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#delete-a-merge-request-note
func (s *NotesService) DeleteMergeRequestNote(pid any, mergeRequest, note int64, options ...RequestOptionFunc) (*Response, error) {
	_, resp, err := do[none](s.client,
		withMethod(http.MethodDelete),
		withPath("projects/%s/merge_requests/%d/notes/%d", ProjectID{pid}, mergeRequest, note),
		withRequestOpts(options...),
	)
	return resp, err
}

// ListEpicNotesOptions represents the available ListEpicNotes() options.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-epic-notes
type ListEpicNotesOptions struct {
	ListOptions
	OrderBy *string `url:"order_by,omitempty" json:"order_by,omitempty"`
	Sort    *string `url:"sort,omitempty" json:"sort,omitempty"`
}

// ListEpicNotes gets a list of all notes for a single epic.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#list-all-epic-notes
func (s *NotesService) ListEpicNotes(gid any, epic int64, opt *ListEpicNotesOptions, options ...RequestOptionFunc) ([]*Note, *Response, error) {
	return do[[]*Note](s.client,
		withPath("groups/%s/epics/%d/notes", GroupID{gid}, epic),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// GetEpicNote returns a single note for an epic.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#get-single-epic-note
func (s *NotesService) GetEpicNote(gid any, epic, note int64, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withPath("groups/%s/epics/%d/notes/%d", GroupID{gid}, epic, note),
		withRequestOpts(options...),
	)
}

// CreateEpicNoteOptions represents the available CreateEpicNote() options.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-epic-note
type CreateEpicNoteOptions struct {
	Body *string `url:"body,omitempty" json:"body,omitempty"`
}

// CreateEpicNote creates a new note for a single merge request.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#create-new-epic-note
func (s *NotesService) CreateEpicNote(gid any, epic int64, opt *CreateEpicNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPost),
		withPath("groups/%s/epics/%d/notes", GroupID{gid}, epic),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// UpdateEpicNoteOptions represents the available UpdateEpicNote() options.
// Will be removed in v5 of the API, use Work Items API instead
//
// GitLab API docs:
// https://docs.gitlab.com/api/notes/#modify-existing-epic-note
type UpdateEpicNoteOptions struct {
	Body *string `url:"body,omitempty" json:"body,omitempty"`
}

// UpdateEpicNote modifies existing note of an epic.
// Will be removed in v5 of the API, use Work Items API instead
//
// https://docs.gitlab.com/api/notes/#modify-existing-epic-note
func (s *NotesService) UpdateEpicNote(gid any, epic, note int64, opt *UpdateEpicNoteOptions, options ...RequestOptionFunc) (*Note, *Response, error) {
	return do[*Note](s.client,
		withMethod(http.MethodPut),
		withPath("groups/%s/epics/%d/notes/%d", GroupID{gid}, epic, note),
		withAPIOpts(opt),
		withRequestOpts(options...),
	)
}

// DeleteEpicNote deletes an existing note of a merge request.
// Will be removed in v5 of the API, use Work Items API instead
//
// https://docs.gitlab.com/api/notes/#delete-an-epic-note
func (s *NotesService) DeleteEpicNote(gid any, epic, note int64, options ...RequestOptionFunc) (*Response, error) {
	_, resp, err := do[none](s.client,
		withMethod(http.MethodDelete),
		withPath("groups/%s/epics/%d/notes/%d", GroupID{gid}, epic, note),
		withRequestOpts(options...),
	)
	return resp, err
}
